"""Injections dependency injection framework

Usage
=====

Declare a class' dependencies::

    @injections.has
    class Hello(object):

        world = injections.depends(World, 'universe')

You must decorate class with ``@injections.has``. All dependencies has a
type(class), which is ``World`` in this case, and a name, which is ``universe``
in this case. This is for having multiple similar dependencies.

.. note::

    Name of dependency is unique among all services in single dependency
    injector and not tied to particular type. This is done to easier support
    subclassing of dependencies (and you can also register subclass with abc
    instead of subclassing directly)

Then at some initialisation code you create dependency injector, and set
apropriate services::

    inj = injections.Container()
    inj['universe'] = World()

Now you can create ``Hello`` instances and inject dependencies to them::

    hello = inj.inject(Hello())
    assert hello.world is inj['universe']

And you can propagate dependencies starting from existing instances::

    h2 = injections.propagate(hello, Hello())
    assert h2.world is hello.world

"""

from .topsort import topologically_sorted


def has(cls):
    """Class decorator that declares dependencies"""
    deps = {}
    for i in dir(cls):
        if i.startswith('__') and i.endswith('__'):
            continue
        val = getattr(cls, i, None)
        if isinstance(val, Dependency):
            deps[i] = val
            if val.name is None:
                val.name = i
    cls.__injections__ = deps
    return cls


class Dependency:
    """Property that represents single dependency"""

    def __init__(self, typ, name=None):
        self.type = typ
        self.name = name

    def __get__(self, inst, owner):
        if inst is None:
            return self
        raise RuntimeError("Dependency {!r} is not configured".format(self))

    def __repr__(self):
        return "<Dependency {!r}:{!r}>".format(self.type, self.name)


depends = Dependency  # nicer declarative name


class Container(object):
    """Container for things that will be dependency-injected

    Set all dependencies with::

        >>> inj = Container()
        >>> inj['name'] = value
        >>> obj = inj.inject(DependentClass())

    Propagate dependencies with::

        >>> injections.propagate(obj, NewObject())
    """

    def __init__(self, *args, **kw):
        self._provides = dict(*args, **kw)

    def inject(self, inst, **renames):
        """Injects dependencies and propagates dependency injector"""
        if renames:
            di = self.clone(**renames)
        else:
            di = self
        pro = di._provides
        inst.__injections_source__ = di
        deps = getattr(inst, '__injections__', None)
        if deps:
            for attr, dep in deps.items():
                val = pro[dep.name]
                if not isinstance(val, dep.type):
                    raise TypeError("Wrong provider for {!r}".format(val))
                setattr(inst, attr, val)
        meth = getattr(inst, '__injected__', None)
        if meth is not None:
            meth()
        return inst

    def __setitem__(self, name, value):
        if name in self._provides:
            raise RuntimeError("Two providers for {!r}".format(name))
        self._provides[name] = value

    def __getitem__(self, name):
        return self._provides[name]

    def __contains__(self, name):
        return name in self._provides

    def clone(self, **renames):
        di = self.__class__()
        mypro = self._provides
        pro = di._provides
        pro.update(mypro)
        for name, alias in renames.items():
            pro[name] = mypro[alias]
        return di

    def interconnect_all(self):
        """Propagate dependencies for provided instances"""
        for dep in topologically_sorted(self._provides):
            if hasattr(dep, '__injections__') and not hasattr(dep, '__injections_source__'):
                self.inject(dep)


def dependencies(cls):
    """Returns dict of dependencies of a class declared with
    ``@injections.has``
    """
    return getattr(cls, '__injections__', {})


def get_container(obj):
    """Returns dependency injector used to construct class

    This instance is useful to propagate dependencies
    """
    try:
        return obj.__injections_source__
    except AttributeError:
        raise RuntimeError("No dependency injector found")


def propagate(source, target):
    """Propagates dependencies from container used to fill `source` to fill
        target's dependencies

         Equivalent of::

            get_container(source).inject(target)
    """
    return get_container(source).inject(target)



from unittest import TestCase

import injections as di


class TestCore(TestCase):

    def setUp(self):
        @di.has
        class Consumer:
            value = di.depends(str, 'name')

            def hello(self):
                return "Hello {}!".format(self.value)

        self.Consumer = Consumer

    def test_ok(self):
        c = di.Container()
        c['name'] = "John"
        cons = c.inject(self.Consumer())
        self.assertEqual(cons.hello(), 'Hello John!')

    def test_interconnect(self):
        @di.has
        class Source:
             value = di.depends(self.Consumer, 'consumer')

             def __injected__(self):
                assert self.value.value is 'John'

        c = di.Container()
        c['name'] = "John"
        c['consumer'] = self.Consumer()
        c.interconnect_all()
        res = c.inject(Source())
        self.assertEqual(res.value.hello(), 'Hello John!')

    def test_cyclic(self):
        class Base:
            pass

        @di.has
        class Processor(Base):
             cache = di.depends(Base, 'cache')

        @di.has
        class Cache(Base):
             processor = di.depends(Base, 'processor')

        c = di.Container()
        c['cache'] = Cache()
        c['processor'] = Processor()
        try:
            c.interconnect_all()
        except RuntimeError as e:
            self.assertSetEqual({'cache', 'processor'},
                set(e.objects_in_cycle.keys()))
        else:
            raise AssertionError("Exception not raised")

    def test_larger_cycle(self):

        @di.has
        class Y(object):
            pass

        @di.has
        class A(object):
             b = di.depends(object, 'b')
             y = di.depends(Y, 'y')

        @di.has
        class B(object):
             c = di.depends(object, 'c')

        @di.has
        class C(object):
             a = di.depends(object, 'a')

        @di.has
        class X(object):
             a = di.depends(object, 'a')

        @di.has
        class Z(object):
             x = di.depends(object, 'x')

        c = di.Container()
        c['a'] = A()
        c['b'] = B()
        c['c'] = C()
        c['x'] = X()
        c['y'] = Y()
        c['z'] = Z()
        try:
            c.interconnect_all()
        except RuntimeError as e:
            self.assertSetEqual({'a', 'b', 'c'},
                set(e.objects_in_cycle.keys()))
        else:
            raise AssertionError("Exception not raised")

    def test_clone(self):
        @di.has
        class User:
            value = di.depends(self.Consumer, 'consumer')

        @di.has
        class Girl:
            father = di.depends(User, 'user')

        c = di.Container()
        c['name'] = "John"
        c['consumer'] = self.Consumer()
        c.interconnect_all()
        c = c.clone()
        c['user'] = User()
        c.interconnect_all()
        r = c.inject(Girl())
        self.assertTrue(r.father is c['user'])
        self.assertTrue(r.father.value is c['consumer'])

    def test_wrong_type(self):
        c = di.Container()
        c['name'] = 1
        with self.assertRaises(TypeError):
            c.inject(self.Consumer())



from collections import defaultdict, deque


class CyclicDependencies(RuntimeError):

    def __init__(self, objects):
        self.objects_in_cycle = objects
        super(CyclicDependencies, self).__init__("The objects {!r} "
                    "are in cyclic dependency. You must resolve them in "
                    "proper order manually.".format(objects))


def topologically_sorted(objects):

    refs = {}
    backrefs = defaultdict(set)
    queue = deque()

    for name, obj in objects.items():
        deps = getattr(obj, '__injections__', None)
        if not deps or getattr(obj, '__injections_source__', None):
            # Already injected objects are like those not having dependencies
            queue.append(name)
            continue
        cur_refs = set(d.name for d in deps.values())
        refs[name] = cur_refs
        for ref in cur_refs:
            backrefs[ref].add(name)
            if name in refs.get(ref, ()):
                raise CyclicDependencies({
                    name: objects[name],
                    ref: objects[ref],
                    })

    while queue:
        obj = queue.popleft()
        yield objects[obj]

        for dep in backrefs.pop(obj, ()):
            cur_refs = refs[dep]
            cur_refs.remove(obj)
            if not cur_refs:
                queue.append(dep)

    if backrefs:
        # There is aready an error, but we should o our best to discover
        # which objects are in cycle, so it's easier to fix by user

        # Basically we reverse dependencies and do same topology sort
        # in the reversed graph we will have same nodes in cycle but we
        # remove nodes that depend on it

        queue = deque(set(refs) - set(backrefs))
        while queue:
            obj = queue.popleft()

            for dep in refs[obj]:
                cur_refs = backrefs[dep]
                cur_refs.remove(obj)
                if not cur_refs:
                    del backrefs[dep]
                    queue.append(dep)

        raise CyclicDependencies({k: objects[k] for k in backrefs})



from .core import (
    Container,
    Dependency,
    has,
    depends,
    propagate,
    get_container,
    dependencies,
    )

__all__ = [
    'Container',
    'Dependency',
    'has',
    'depends',
    'propagate',
    'get_container',
    'dependencies',
    ]



#!/usr/bin/env python

from distutils.core import setup

setup(name='injections',
      version='0.2.1',
      description='Simple dependency injection library',
      author='Paul Colomiets',
      author_email='paul@colomiets.name',
      url='http://github.com/tailhook/injections',
      packages=[
        'injections',
        ],
      classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 2',
        ],
     )
